@@ -4,7 +4,19 @@ use 5.006;
 use Carp;
 use strict;
 use warnings;
-use IO::Socket;
+
+# modified by M.Vincent for IPv6 support
+use Socket qw(AF_INET AF_INET6 SO_ERROR);
+my $HAVE_IO_Socket_IP = 0;
+eval "use IO::Socket::IP -register";
+if (!$@)
+{
+	$HAVE_IO_Socket_IP = 1;
+}
+else
+{
+	eval "use IO::Socket::INET";
+}
 
 require Exporter;
 
@@ -86,7 +98,7 @@ our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );
 
 our @EXPORT = qw( );
 
-our $VERSION = '0.03';
+our $VERSION = '0.05';
 
 our $LASTERROR;
 
@@ -108,53 +120,106 @@ sub new
 	# setting defaults
 	$cfg{'FileName'} or $cfg{'RootDir'} or croak "Usage: \$tftpdOBJ = Net::TFTPd->new(['RootDir' => 'path/to/files' | 'FileName' => 'path/to/file'] [, [ LocalPort => portnum ] [, ...]] );";
 
-	if($cfg{'RootDir'} and not -d($cfg{'RootDir'}) )
+	if ($cfg{'RootDir'} and not -d($cfg{'RootDir'}) )
 	{
 		$LASTERROR = sprintf 'RootDir \'%s\' not found or is not a valid directory name\n', $cfg{'RootDir'};
-		return(undef);
+		return (undef);
 	}
 
-	if($cfg{'FileName'} and not -e($cfg{'FileName'}) )
+	if ($cfg{'FileName'} and not -e($cfg{'FileName'}) )
 	{
 		$LASTERROR = sprintf 'FileName \'%s\' not found or is not a valid filename\n', $cfg{'FileName'};
-		return(undef);
+		return (undef);
 	}
 
 	my %params = (
 		'Proto' => 'udp',
-		'LocalPort' => $cfg{'LocalPort'} || TFTP_DEFAULT_PORT,
+		'LocalPort' => $cfg{'LocalPort'} || TFTP_DEFAULT_PORT
 	);
 
+	# modified by M.Vincent for IPv6 support
+	if (defined($cfg{'Family'}))
+	{
+		if ($cfg{'Family'} =~ /^(?:(?:(:?ip)?v?(?:4|6))|${\AF_INET}|${\AF_INET6})$/)
+		{
+			if ($cfg{'Family'} =~ /^(?:(?:(:?ip)?v?4)|${\AF_INET})$/)
+			{
+				$params{'Family'} = AF_INET;
+			}
+			else
+			{
+				if (!$HAVE_IO_Socket_IP)
+				{
+					$LASTERROR = "IO::Socket::IP required for IPv6";
+					return (undef);
+				}
+				$params{'Family'} = AF_INET6;
+			}
+		}
+		else
+		{
+			$LASTERROR = "Invalid family - $cfg{'Family'}";
+			return (undef);
+		}
+	}
+	else
+	{
+		$params{'Family'} = AF_INET;
+	}
+	
 	# bind only to specified address
-	if($cfg{'LocalAddr'})
+	if ($cfg{'LocalAddr'})
 	{
 		$params{'LocalAddr'} = $cfg{'LocalAddr'};
 	}
 
-	if(my $udpserver = IO::Socket::INET->new(%params))
+	if ($HAVE_IO_Socket_IP)
 	{
-#removed for using this module with IO v. 1.2301 under SUSE 10.1, O.Z. 15.08.2007
-#		$udpserver->setsockopt(SOL_SOCKET, SO_RCVBUF, 0);
-#		$udpserver->setsockopt(SOL_SOCKET, SO_SNDBUF, 0);
-
-		return bless {
-			'LocalPort'   => TFTP_DEFAULT_PORT,
-			'Timeout'     => 10,
-			'ACKtimeout'  => 4,
-			'ACKretries'  => 4,
-			'Readable'    => 1,
-			'Writable'    => 0,
-			'CallBack'    => undef,
-			'BlkSize'     => TFTP_DEFAULT_BLKSIZE,
-			'Debug'       => 0,
-			%cfg,         # merge user parameters
-			'_UDPSERVER_' => $udpserver
-		}, $class;
+		if (my $udpserver = IO::Socket::IP->new(%params))
+		{
+			return bless {
+				'LocalPort'   => TFTP_DEFAULT_PORT,
+				'Timeout'     => 10,
+				'ACKtimeout'  => 4,
+				'ACKretries'  => 4,
+				'Readable'    => 1,
+				'Writable'    => 0,
+				'CallBack'    => undef,
+				'BlkSize'     => TFTP_DEFAULT_BLKSIZE,
+				'Debug'       => 0,
+				%cfg,         # merge user parameters
+				'_UDPSERVER_' => $udpserver
+			}, $class;
+		}
+		else
+		{
+			$LASTERROR = "Error opening socket for listener: $@\n";
+			return (undef);
+		}
 	}
 	else
 	{
-		$LASTERROR = "Error opening socket for listener: $@\n";
-		return(undef);
+		if (my $udpserver = IO::Socket::INET->new(%params))
+		{
+			return bless {
+				'LocalPort'   => TFTP_DEFAULT_PORT,
+				'Timeout'     => 10,
+				'ACKtimeout'  => 4,
+				'ACKretries'  => 4,
+				'Readable'    => 1,
+				'Writable'    => 0,
+				'CallBack'    => undef,
+				'BlkSize'     => TFTP_DEFAULT_BLKSIZE,
+				'Debug'       => 0,
+				%cfg,         # merge user parameters
+				'_UDPSERVER_' => $udpserver
+			}, $class;
+		}
+		else
+		{
+			$LASTERROR = "Error opening socket for listener: $@\n";
+			return (undef);
+		}
 	}
 }
 
@@ -195,7 +260,7 @@ sub waitRQ
 	if (select($rout=$rin, undef, $eout=$ein, $Timeout))
 	{
 		# read the message
-		if($udpserver->recv($datagram, TFTP_MAX_BLKSIZE + 4))
+		if ($udpserver->recv($datagram, TFTP_MAX_BLKSIZE + 4))
 		{
 			# decode the message
 			($opcode, $datain) = unpack("na*", $datagram);
@@ -203,9 +268,8 @@ sub waitRQ
 			$request->{'_REQUEST_'}{'OPCODE'} = $opcode;
 
 			# get peer port and address
-			my($peerport, $peeraddr) = sockaddr_in($udpserver->peername);
-			$request->{'_REQUEST_'}{'PeerPort'} = $peerport;
-			$request->{'_REQUEST_'}{'PeerAddr'} = inet_ntoa($peeraddr);
+			$request->{'_REQUEST_'}{'PeerPort'} = $udpserver->peerport;
+			$request->{'_REQUEST_'}{'PeerAddr'} = $udpserver->peerhost;
 
 			# get filename and transfer mode
 			my @datain = split("\0", $datain);
@@ -215,8 +279,10 @@ sub waitRQ
 			$request->{'_REQUEST_'}{'BlkSize'} = TFTP_DEFAULT_BLKSIZE;
 			$request->{'_REQUEST_'}{'LASTACK'} = 0;
 			$request->{'_REQUEST_'}{'PREVACK'} = -1;
+			# counter for transferred bytes
+			$request->{'_REQUEST_'}{'TotalBytes'} = 0;
 
-			if(scalar(@datain) >= 2)
+			if (scalar(@datain) >= 2)
 			{
 				$request->{'_REQUEST_'}{'RFC2347'} = { @datain };
 			}
@@ -227,13 +293,13 @@ sub waitRQ
 		{
 			$! = $udpserver->sockopt(SO_ERROR);
 			$LASTERROR = sprintf "Socket RECV error: %s\n", $!;
-			return(undef);
+			return (undef);
 		}
 	}
 	else
 	{
 		$LASTERROR = "Timed out waiting for RRQ/WRQ";
-		return(0);
+		return (0);
 	}
 }
 
@@ -246,57 +312,58 @@ sub processRQ
 	# the request object
 	my $self = shift;
 
-	if(defined($self->newSOCK()))
+	if (defined($self->newSOCK()))
 	{
-		if($self->{'_REQUEST_'}{'Mode'} ne 'OCTET')
+		# modified for supporting NETASCII transfers on 25/05/2009
+		if (($self->{'_REQUEST_'}{'Mode'} ne 'OCTET') && ($self->{'_REQUEST_'}{'Mode'} ne 'NETASCII'))
 		{
 			#request is not OCTET
 			$LASTERROR = sprintf "%s transfer mode is not supported\n", $self->{'_REQUEST_'}{'Mode'};
 			$self->sendERR(0, $LASTERROR);
-			return(undef);
+			return (undef);
 		}
 
 		# new socket opened successfully
-		if($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_RRQ)
+		if ($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_RRQ)
 		{
 			#################
 			# opcode is RRQ #
 			#################
-			if($self->{'Readable'})
+			if ($self->{'Readable'})
 			{
 				# read is permitted
-				if($self->{'_REQUEST_'}{'FileName'} =~ /\.\.[\\\/]/)
+				if ($self->{'_REQUEST_'}{'FileName'} =~ /\.\.[\\\/]/)
 				{
 					# requested file contains '..\' or '../'
 					$LASTERROR = sprintf 'Access to \'%s\' is not permitted to %s', $self->{'_REQUEST_'}{'FileName'}, $self->{'_REQUEST_'}{'PeerAddr'};
 					$self->sendERR(2);
-					return(undef);
+					return (undef);
 				}
 
-				if(defined($self->checkFILE()))
+				if (defined($self->checkFILE()))
 				{
 					# file is present
-					if(defined($self->negotiateOPTS()))
+					if (defined($self->negotiateOPTS()))
 					{
 						# RFC 2347 options negotiated
-						if(defined($self->openFILE()))
+						if (defined($self->openFILE()))
 						{
 							# file opened for read, start the transfer
-							if(defined($self->sendFILE()))
+							if (defined($self->sendFILE()))
 							{
 								# file sent successfully
-								return(1);
+								return (1);
 							}
 							else
 							{
 								# error sending file
-								return(undef);
+								return (undef);
 							}
 						}
 						else
 						{
 							# error opening file
-							return(undef);
+							return (undef);
 						}
 					}
 					else
@@ -304,7 +371,7 @@ sub processRQ
 						# error negotiating options
 						$LASTERROR = "TFTP error 8: Option negotiation\n";
 						$self->sendERR(8);
-						return(undef);
+						return (undef);
 					}
 				}
 				else
@@ -312,7 +379,7 @@ sub processRQ
 					# file not found
 					$LASTERROR = sprintf 'File \'%s\' not found', $self->{'_REQUEST_'}{'FileName'};
 					$self->sendERR(1);
-					return(undef);
+					return (undef);
 				}
 			}
 			else
@@ -320,43 +387,43 @@ sub processRQ
 				# if server is not readable
 				$LASTERROR = "TFTP Error: Access violation";
 				$self->sendERR(2);
-				return(undef);
+				return (undef);
 			}
 		}
-		elsif($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_WRQ)
+		elsif ($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_WRQ)
 		{
 			#################
 			# opcode is WRQ #
 			#################
-			if($self->{'Writable'})
+			if ($self->{'Writable'})
 			{
 				# write is permitted
-				if($self->{'_REQUEST_'}{'FileName'} =~ /\.\.[\\\/]/)
+				if ($self->{'_REQUEST_'}{'FileName'} =~ /\.\.[\\\/]/)
 				{
 					# requested file contains '..\' or '../'
 					$LASTERROR = sprintf 'Access to \'%s\' is not permitted to %s', $self->{'_REQUEST_'}{'FileName'}, $self->{'_REQUEST_'}{'PeerAddr'};
 					$self->sendERR(2);
-					return(undef);
+					return (undef);
 				}
 
-				if(!defined($self->checkFILE()))
+				if (!defined($self->checkFILE()))
 				{
 					# RFC 2347 options negotiated
-					if(defined($self->openFILE()))
+					if (defined($self->openFILE()))
 					{
 						# file is not present
-						if(defined($self->negotiateOPTS()))
+						if (defined($self->negotiateOPTS()))
 						{
 							# file opened for write, start the transfer
-							if(defined($self->recvFILE()))
+							if (defined($self->recvFILE()))
 							{
 								# file received successfully
-								return(1);
+								return (1);
 							}
 							else
 							{
 								# error receiving file
-								return(undef);
+								return (undef);
 							}
 						}
 						else
@@ -364,14 +431,14 @@ sub processRQ
 							# error negotiating options
 							$LASTERROR = "TFTP error 8: Option negotiation\n";
 							$self->sendERR(8);
-							return(undef);
+							return (undef);
 						}
 					}
 					else
 					{
 						# error opening file
 						$self->sendERR(3);
-						return(undef);
+						return (undef);
 					}
 				}
 				else
@@ -379,7 +446,7 @@ sub processRQ
 					# file not found
 					$LASTERROR = sprintf 'File \'%s\' already exists', $self->{'_REQUEST_'}{'FileName'};
 					$self->sendERR(6);
-					return(undef);
+					return (undef);
 				}
 			}
 			else
@@ -387,7 +454,7 @@ sub processRQ
 				# if server is not writable
 				$LASTERROR = "TFTP Error: Access violation";
 				$self->sendERR(2);
-				return(undef);
+				return (undef);
 			}
 		}
 		else
@@ -397,15 +464,86 @@ sub processRQ
 			#################
 			$LASTERROR = sprintf "Opcode %d not supported as request", $self->{'_REQUEST_'}{'OPCODE'};
 			$self->sendERR(4);
-			return(undef);
+			return (undef);
 		}
 	}
 	else
 	{
-		return(undef);
+		return (undef);
 	}
 }
 
+#
+# Usage: $requestOBJ->getTotalBytes();
+# returns the number of bytes transferred by the request
+#
+sub getTotalBytes
+{
+	# the request object
+	my $self = shift;
+	
+	return $self->{'_REQUEST_'}{'TotalBytes'};
+}
+
+#
+# Usage: $requestOBJ->getFileName();
+# returns the requested file name
+#
+sub getFileName
+{
+	# the request object
+	my $self = shift;
+	
+	return $self->{'_REQUEST_'}{'FileName'};
+}
+
+#
+# Usage: $requestOBJ->getMode();
+# returns the transfer mode for the request
+#
+sub getMode
+{
+	# the request object
+	my $self = shift;
+	
+	return $self->{'_REQUEST_'}{'Mode'};
+}
+
+#
+# Usage: $requestOBJ->getPeerAddr();
+# returns the address of the requesting client
+#
+sub getPeerAddr
+{
+	# the request object
+	my $self = shift;
+	
+	return $self->{'_REQUEST_'}{'PeerAddr'};
+}
+
+#
+# Usage: $requestOBJ->getPeerPort();
+# returns the port of the requesting client
+#
+sub getPeerPort
+{
+	# the request object
+	my $self = shift;
+	
+	return $self->{'_REQUEST_'}{'PeerPort'};
+}
+
+#
+# Usage: $requestOBJ->getBlkSize();
+# returns the block size used for the transfer
+#
+sub getBlkSize
+{
+	# the request object
+	my $self = shift;
+	
+	return $self->{'_REQUEST_'}{'BlkSize'};
+}
 
 #
 # Usage: $requestOBJ->newSOCK();
@@ -424,27 +562,39 @@ sub newSOCK
 	);
 
 	# bind only to specified address
-	if($self->{'Address'})
+	if ($self->{'Address'})
 	{
 		$params{'LocalAddr'} = $self->{'Address'};
 	}
 
 	# open socket
-	if(my $udpserver = IO::Socket::INET->new(%params))
+	if ($HAVE_IO_Socket_IP)
 	{
-#removed for using this module with IO v. 1.2301 under SUSE 10.1, O.Z. 15.08.2007
-#		$udpserver->setsockopt(SOL_SOCKET, SO_RCVBUF, 0);
-#		$udpserver->setsockopt(SOL_SOCKET, SO_SNDBUF, 0);
-
-		$self->{'_UDPSERVER_'} = $udpserver;
-		return(1);
+		if (my $udpserver = IO::Socket::IP->new(%params))
+		{
+			$self->{'_UDPSERVER_'} = $udpserver;
+			return (1);
+		}
+		else
+		{
+			$LASTERROR = "Error opening socket for reply: $@\n";
+			return (undef);
+		}
 	}
 	else
 	{
-		$LASTERROR = "Error opening socket for reply: $@\n";
-		return(undef);
+		if (my $udpserver = IO::Socket::INET->new(%params))
+		{
+			$self->{'_UDPSERVER_'} = $udpserver;
+			return (1);
+		}
+		else
+		{
+			$LASTERROR = "Error opening socket for reply: $@\n";
+			return (undef);
+		}
 	}
-}
+}	
 
 
 #
@@ -456,15 +606,15 @@ sub negotiateOPTS
 	# the request object
 	my $self = shift;
 
-	if($self->{'_REQUEST_'}{'RFC2347'})
+	if ($self->{'_REQUEST_'}{'RFC2347'})
 	{
 		# parse RFC 2347 options if present
 		foreach my $option (keys(%{ $self->{'_REQUEST_'}{'RFC2347'} }))
 		{
-			if(uc($option) eq 'BLKSIZE')
+			if (uc($option) eq 'BLKSIZE')
 			{
 				# Negotiate the blocksize
-				if($self->{'_REQUEST_'}{'RFC2347'}{$option} > TFTP_MAX_BLKSIZE or $self->{'_REQUEST_'}{'RFC2347'}{$option} < TFTP_MIN_BLKSIZE)
+				if ($self->{'_REQUEST_'}{'RFC2347'}{$option} > TFTP_MAX_BLKSIZE or $self->{'_REQUEST_'}{'RFC2347'}{$option} < TFTP_MIN_BLKSIZE)
 				{
 					$self->{'_REQUEST_'}{'RFC2347'}{$option} = $self->{'BlkSize'};
 				}
@@ -474,10 +624,10 @@ sub negotiateOPTS
 					$self->{'BlkSize'} = $self->{'_RESPONSE_'}{'RFC2347'}{$option};
 				}
 			}
-			elsif(uc($option) eq 'TSIZE')
+			elsif (uc($option) eq 'TSIZE')
 			{
 				# Negotiate the transfer size
-				if($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_RRQ)
+				if ($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_RRQ)
 				{
 					$self->{'_RESPONSE_'}{'RFC2347'}{$option} = $self->{'FileSize'};
 				}
@@ -486,10 +636,10 @@ sub negotiateOPTS
 					$self->{'FileSize'} = $self->{'_REQUEST_'}{'RFC2347'}{$option};
 				}
 			}
-			elsif(uc($option) eq 'TIMEOUT')
+			elsif (uc($option) eq 'TIMEOUT')
 			{
 				# Negotiate the transfer timeout
-				if($self->{'_REQUEST_'}{'RFC2347'}{$option} > TFTP_MAX_TIMEOUT or $self->{'_REQUEST_'}{'RFC2347'}{$option} < TFTP_MIN_TIMEOUT)
+				if ($self->{'_REQUEST_'}{'RFC2347'}{$option} > TFTP_MAX_TIMEOUT or $self->{'_REQUEST_'}{'RFC2347'}{$option} < TFTP_MIN_TIMEOUT)
 				{
 					$self->{'_RESPONSE_'}{'RFC2347'}{$option} = $self->{'ACKtimeout'};
 				}
@@ -506,36 +656,36 @@ sub negotiateOPTS
 		}
 
 		# post processing
-		if($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_WRQ)
+		if ($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_WRQ)
 		{
-			if($self->{'FileSize'} and $self->{'BlkSize'})
+			if ($self->{'FileSize'} and $self->{'BlkSize'})
 			{
 				$self->{'_REQUEST_'}{'LASTACK'} = int($self->{'FileSize'} / $self->{'BlkSize'}) + 1;
 			}
 		}
 
 		# send OACK for RFC 2347 options
-		return($self->sendOACK());
+		return ($self->sendOACK());
 	}
 	else
 	{
-		if($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_WRQ)
+		if ($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_WRQ)
 		{
 			# opcode is WRQ: send ACK for datablock 0
-			if($self->{'_UDPSERVER_'}->send(pack("nn", TFTP_OPCODE_ACK, 0)))
+			if ($self->{'_UDPSERVER_'}->send(pack("nn", TFTP_OPCODE_ACK, 0)))
 			{
-				return(1);
+				return (1);
 			}
 			else
 			{
 				$! = $self->{'_UDPSERVER_'}->sockopt(SO_ERROR);
 				$LASTERROR = sprintf "Socket SEND error: %s\n", $!;
-				return(undef);
+				return (undef);
 			}
 		}
 		else
 		{
-			return(1);
+			return (1);
 		}
 	}
 }
@@ -550,25 +700,27 @@ sub readFILE
 	my $self = shift;
 	my $datablk = shift;
 
-	if($self->{'_REQUEST_'}{'PREVACK'} < $self->{'_REQUEST_'}{'LASTACK'})
+	if ($self->{'_REQUEST_'}{'PREVACK'} < $self->{'_REQUEST_'}{'LASTACK'})
 	{
 		# if requested block is next block, read next block and return bytes read
 		my $fh = $self->{'_REQUEST_'}{'_FH_'};
-		my $bytes = read($fh, $$datablk, $self->{'BlkSize'});
-		if(defined($bytes))
+		# modified for supporting NETASCII transfers on 25/05/2009
+		# my $bytes = read ($fh, $$datablk, $self->{'BlkSize'});
+		my $bytes = sysread($fh, $$datablk, $self->{'BlkSize'});
+		if (defined($bytes))
 		{
-			return($bytes);
+			return ($bytes);
 		}
 		else
 		{
 			$LASTERROR = sprintf "Error $! reading file '%s'", $self->{'_REQUEST_'}{'FileName'};
-			return(undef);
+			return (undef);
 		}
 	}
 	else
 	{
 		# if requested block is last block, return length of last block
-		return(length($$datablk));
+		return (length($$datablk));
 	}
 }
 
@@ -582,23 +734,23 @@ sub writeFILE
 	my $self = shift;
 	my $datablk = shift;
 
-	if($self->{'_REQUEST_'}{'PREVBLK'} > $self->{'_REQUEST_'}{'LASTBLK'})
+	if ($self->{'_REQUEST_'}{'PREVBLK'} > $self->{'_REQUEST_'}{'LASTBLK'})
 	{
 		# if last block is < than previous block, return length of last block
-		return(length($$datablk));
+		return (length($$datablk));
 	}
-	elsif($self->{'_REQUEST_'}{'LASTBLK'} eq ($self->{'_REQUEST_'}{'PREVBLK'} + 1))
+	elsif ($self->{'_REQUEST_'}{'LASTBLK'} eq ($self->{'_REQUEST_'}{'PREVBLK'} + 1))
 	{
 		# if block is next block, write next block and return bytes written
 		my $fh = $self->{'_REQUEST_'}{'_FH_'};
 		my $bytes = syswrite($fh, $$datablk);
-		return($bytes);
+		return ($bytes);
 	}
 	else
 	{
 		$LASTERROR = sprintf "TFTP Error DATA block %d is out of sequence, expected block was %d", $self->{'_REQUEST_'}{'LASTBLK'}, $self->{'_REQUEST_'}{'PREVBLK'} + 1;
 		$self->sendERR(5);
-		return(undef);
+		return (undef);
 	}
 }
 
@@ -611,18 +763,20 @@ sub sendFILE
 {
 	my $self = shift;
 
-	while(1)
+	while (1)
 	{
-		if($self->{'_REQUEST_'}{'LASTACK'} < $self->{'_REQUEST_'}{'LASTBLK'})
+		if ($self->{'_REQUEST_'}{'LASTACK'} < $self->{'_REQUEST_'}{'LASTBLK'})
 		{
 			my $datablk = 0;
-			if(defined($self->readFILE(\$datablk)))
+			if (defined($self->readFILE(\$datablk)))
 			{
 				# read from file successful
-				if($self->sendDATA(\$datablk))
+				# increment the transferred bytes counter
+				$self->{'_REQUEST_'}{'TotalBytes'} += length($datablk);
+				if ($self->sendDATA(\$datablk))
 				{
 					# send to socket successful
-					if($self->{'CallBack'})
+					if ($self->{'CallBack'})
 					{
 						&{$self->{'CallBack'}}($self);
 					}
@@ -630,19 +784,19 @@ sub sendFILE
 				else
 				{
 					# error sending to socket
-					return(undef);
+					return (undef);
 				}
 			}
 			else
 			{
 				# error reading from file
-				return(undef);
+				return (undef);
 			}
 		}
 		else
 		{
 			# transfer completed
-			return(1);
+			return (1);
 		}
 	}
 }
@@ -659,23 +813,25 @@ sub recvFILE
 	$self->{'_REQUEST_'}{'LASTBLK'} = 0;
 	$self->{'_REQUEST_'}{'PREVBLK'} = 0;
 
-	while(1)
+	while (1)
 	{
 		my $datablk = 0;
-		if($self->recvDATA(\$datablk))
+		if ($self->recvDATA(\$datablk))
 		{
 			# DATA received
-			if(defined($self->writeFILE(\$datablk)))
+			if (defined($self->writeFILE(\$datablk)))
 			{
 				# DATA written to file
 				my $udpserver = $self->{'_UDPSERVER_'};
 
-				if(defined($udpserver->send(pack("nn", TFTP_OPCODE_ACK, $self->{'_REQUEST_'}{'LASTBLK'}))))
+				if (defined($udpserver->send(pack("nn", TFTP_OPCODE_ACK, $self->{'_REQUEST_'}{'LASTBLK'}))))
 				{
 					# sent ACK
-					if(length($datablk) < $self->{'BlkSize'})
+					# increment the transferred bytes counter
+					$self->{'_REQUEST_'}{'TotalBytes'} += length($datablk);
+					if (length($datablk) < $self->{'BlkSize'})
 					{
-						return(1);
+						return (1);
 					}
 					else
 					{
@@ -686,19 +842,19 @@ sub recvFILE
 				{
 					$! = $udpserver->sockopt(SO_ERROR);
 					$LASTERROR = sprintf "Socket SEND error: %s\n", $!;
-					return(undef);
+					return (undef);
 				}
 			}
 			else
 			{
 				# error writing data
-				return(undef);
+				return (undef);
 			}
 		}
 		else
 		{
 			# timeout waiting for data
-			return(undef);
+			return (undef);
 		}
 	}
 }
@@ -721,14 +877,14 @@ sub recvDATA
 	vec($rin, fileno($udpserver), 1) = 1;
 
 	# wait for data
-	if(select($rout=$rin, undef, $eout=$ein, $self->{'ACKtimeout'}))
+	if (select($rout=$rin, undef, $eout=$ein, $self->{'ACKtimeout'}))
 	{
 		# read the message
-		if($udpserver->recv($datagram, $self->{'BlkSize'} + 4))
+		if ($udpserver->recv($datagram, $self->{'BlkSize'} + 4))
 		{
 			# decode the message
 			($opcode, $datain) = unpack("na*", $datagram);
-			if($opcode eq TFTP_OPCODE_DATA)
+			if ($opcode eq TFTP_OPCODE_DATA)
 			{
 				# message is DATA
 				$self->{'_REQUEST_'}{'PREVBLK'} = $self->{'_REQUEST_'}{'LASTBLK'};
@@ -739,32 +895,32 @@ sub recvDATA
 					&{$self->{'CallBack'}}($self);
 				}
 
-				return(1);
+				return (1);
 			}
-			elsif($opcode eq TFTP_OPCODE_ERROR)
+			elsif ($opcode eq TFTP_OPCODE_ERROR)
 			{
 				# message is ERR
 				$LASTERROR = sprintf "TFTP error message: %s", $datain;
-				return(undef);
+				return (undef);
 			}
 			else
 			{
 				# other messages...
 				$LASTERROR = sprintf "Opcode %d not supported waiting for DATA\n", $opcode;
-				return(undef);
+				return (undef);
 			}
 		}
 		else
 		{
 			$! = $udpserver->sockopt(SO_ERROR);
 			$LASTERROR = sprintf "Socket RECV error: %s\n", $!;
-			return(undef);
+			return (undef);
 		}
 	}
 	else
 	{
 		$LASTERROR = sprintf "Timeout occurred on DATA packet %d\n", $self->{'_REQUEST_'}{'LASTBLK'} + 1;
-		return(undef);
+		return (undef);
 	}
 }
 
@@ -783,50 +939,50 @@ sub sendDATA
 
 	my ($datagram, $opcode, $datain);
 
-	while($retry < $self->{'ACKretries'})
+	while ($retry < $self->{'ACKretries'})
 	{
-		if($udpserver->send(pack("nna*", TFTP_OPCODE_DATA, $self->{'_REQUEST_'}{'LASTACK'} + 1, $$datablk)))
+		if ($udpserver->send(pack("nna*", TFTP_OPCODE_DATA, $self->{'_REQUEST_'}{'LASTACK'} + 1, $$datablk)))
 		{
 			# vars for IO select
 			my ($rin, $rout, $ein, $eout) = ('', '', '', '');
 			vec($rin, fileno($udpserver), 1) = 1;
 
 			# wait for acknowledge
-			if(select($rout=$rin, undef, $eout=$ein, $self->{'ACKtimeout'}))
+			if (select($rout=$rin, undef, $eout=$ein, $self->{'ACKtimeout'}))
 			{
 				# read the message
-				if($udpserver->recv($datagram, TFTP_MAX_BLKSIZE + 4))
+				if ($udpserver->recv($datagram, TFTP_MAX_BLKSIZE + 4))
 				{
 					# decode the message
 					($opcode, $datain) = unpack("na*", $datagram);
-					if($opcode eq TFTP_OPCODE_ACK)
+					if ($opcode eq TFTP_OPCODE_ACK)
 					{
 						# message is ACK
                   # modified for supporting more blocks count than 65535, O.Z. 15.08.2007
 						$self->{'_REQUEST_'}{'PREVACK'} = $self->{'_REQUEST_'}{'LASTACK'};
-                  if(int(($self->{'_REQUEST_'}{'LASTACK'}+1) % 65536) == unpack("n", $datain)){
+                  if (int(($self->{'_REQUEST_'}{'LASTACK'}+1) % 65536) == unpack("n", $datain)){
                     $self->{'_REQUEST_'}{'LASTACK'}++;
                   };
-						return(1);
+						return (1);
 					}
-					elsif($opcode eq TFTP_OPCODE_ERROR)
+					elsif ($opcode eq TFTP_OPCODE_ERROR)
 					{
 						# message is ERR
 						$LASTERROR = sprintf "TFTP error message: %s", $datain;
-						return(undef);
+						return (undef);
 					}
 					else
 					{
 						# other messages...
 						$LASTERROR = sprintf "Opcode %d not supported as a reply to DATA\n", $opcode;
-						return(undef);
+						return (undef);
 					}
 				}
 				else
 				{
 					$! = $udpserver->sockopt(SO_ERROR);
 					$LASTERROR = sprintf "Socket RECV error: %s\n", $!;
-					return(undef);
+					return (undef);
 				}
 			}
 			else
@@ -840,7 +996,7 @@ sub sendDATA
 		{
 			$! = $udpserver->sockopt(SO_ERROR);
 			$LASTERROR = sprintf "Socket SEND error: %s\n", $!;
-			return(undef);
+			return (undef);
 		}
 	}
 }
@@ -854,15 +1010,15 @@ sub openFILE
 	# the request object
 	my $self = shift;
 
-	if($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_RRQ)
+	if ($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_RRQ)
 	{
 		########################################
 		# opcode is RRQ, open file for reading #
 		########################################
-		if(open(RFH, "<".$self->{'_REQUEST_'}{'FileName'}))
+		if (open(RFH, "<".$self->{'_REQUEST_'}{'FileName'}))
 		{
 			# if OCTET mode, set FileHandle to binary mode...
-			if($self->{'_REQUEST_'}{'Mode'} eq 'OCTET')
+			if ($self->{'_REQUEST_'}{'Mode'} eq 'OCTET')
 			{
 				binmode(RFH);
 			}
@@ -873,23 +1029,23 @@ sub openFILE
 			# save the filehandle reference...
 			$self->{'_REQUEST_'}{'_FH_'} = *RFH;
 
-			return(1);
+			return (1);
 		}
 		else
 		{
 			$LASTERROR = sprintf "Error opening file \'%s\' for reading\n", $self->{'_REQUEST_'}{'FileName'};
-			return(undef);
+			return (undef);
 		}
 	}
-	elsif($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_WRQ)
+	elsif ($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_WRQ)
 	{
 		########################################
 		# opcode is WRQ, open file for writing #
 		########################################
-		if(open(WFH, ">".$self->{'_REQUEST_'}{'FileName'}))
+		if (open(WFH, ">".$self->{'_REQUEST_'}{'FileName'}))
 		{
 			# if OCTET mode, set FileHandle to binary mode...
-			if($self->{'_REQUEST_'}{'Mode'} eq 'OCTET')
+			if ($self->{'_REQUEST_'}{'Mode'} eq 'OCTET')
 			{
 				binmode(WFH);
 			}
@@ -897,12 +1053,12 @@ sub openFILE
 			# save the filehandle reference...
 			$self->{'_REQUEST_'}{'_FH_'} = *WFH;
 
-			return(1);
+			return (1);
 		}
 		else
 		{
 			$LASTERROR = sprintf "Error opening file \'%s\' for writing\n", $self->{'_REQUEST_'}{'FileName'};
-			return(undef);
+			return (undef);
 		}
 	}
 	else
@@ -911,7 +1067,7 @@ sub openFILE
 		# other opcodes are errors #
 		############################
 		$LASTERROR = sprintf "OPCODE %d is not supported\n", $self->{'_REQUEST_'}{'OPCODE'};
-		return(undef);
+		return (undef);
 	}
 }
 
@@ -923,21 +1079,21 @@ sub closeFILE
 {
 	my $self = shift;
 
-	if($self->{'_REQUEST_'}{'_FH_'})
+	if ($self->{'_REQUEST_'}{'_FH_'})
 	{
-		if(close($self->{'_REQUEST_'}{'_FH_'}))
+		if (close($self->{'_REQUEST_'}{'_FH_'}))
 		{
-			return(1);
+			return (1);
 		}
 		else
 		{
 			$LASTERROR = "Error closing filehandle\n";
-			return(undef);
+			return (undef);
 		}
 	}
 	else
 	{
-		return(1);
+		return (1);
 	}
 }
 
@@ -953,33 +1109,33 @@ sub checkFILE
 	# requested file
 	my $reqfile = $self->{'_REQUEST_'}{'FileName'};
 
-	if($self->{'FileName'})
+	if ($self->{'FileName'})
 	{
 		# filename is fixed
 		$self->{'_REQUEST_'}{'FileName'} = $self->{'FileName'};
 
-		if(($self->{'FileName'} =~ /$reqfile/) and -e($self->{'FileName'}))
+		if (($self->{'FileName'} =~ /$reqfile/) and -e($self->{'FileName'}))
 		{
 			# fixed name contains requested file and file exists
 			$self->{'FileSize'} = -s($self->{'FileName'});
-			return(1);
+			return (1);
 		}
 	}
-	elsif($self->{'RootDir'})
+	elsif ($self->{'RootDir'})
 	{
 		# rootdir is fixed
 		$reqfile = $self->{'RootDir'}.'/'.$reqfile;
 		$self->{'_REQUEST_'}{'FileName'} = $reqfile;
 
-		if(-e($reqfile))
+		if (-e($reqfile))
 		{
 			# file exists in rootdir
 			$self->{'FileSize'} = -s($reqfile);
-			return(1);
+			return (1);
 		}
 	}
 
-	return(undef);
+	return (undef);
 }
 
 #
@@ -995,50 +1151,50 @@ sub sendOACK
 
 	my ($datagram, $opcode, $datain);
 
-	while($retry < $self->{'ACKretries'})
+	while ($retry < $self->{'ACKretries'})
 	{
 		# send oack
 		my $data = join("\0", %{ $self->{'_RESPONSE_'}{'RFC2347'} })."\0";
-		if($udpserver->send(pack("na*", TFTP_OPCODE_OACK, $data)))
+		if ($udpserver->send(pack("na*", TFTP_OPCODE_OACK, $data)))
 		{
 			# opcode is RRQ
-			if($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_RRQ)
+			if ($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_RRQ)
 			{
 				# vars for IO select
 				my ($rin, $rout, $ein, $eout) = ('', '', '', '');
 				vec($rin, fileno($udpserver), 1) = 1;
 
 				# wait for acknowledge
-				if(select($rout=$rin, undef, $eout=$ein, $self->{'ACKtimeout'}))
+				if (select($rout=$rin, undef, $eout=$ein, $self->{'ACKtimeout'}))
 				{
 					# read the message
-					if($udpserver->recv($datagram, TFTP_MAX_BLKSIZE + 4))
+					if ($udpserver->recv($datagram, TFTP_MAX_BLKSIZE + 4))
 					{
 						# decode the message
 						($opcode, $datain) = unpack("na*", $datagram);
-						if($opcode == TFTP_OPCODE_ACK)
+						if ($opcode == TFTP_OPCODE_ACK)
 						{
 							# message is ACK
 							my $lastack = unpack("n", $datain);
-							if($lastack)
+							if ($lastack)
 							{
 								# ack is not for block 0... ERROR
 								$LASTERROR = sprintf "Received ACK for block %d instead of 0", $lastack;
-								return(undef);
+								return (undef);
 							}
 							return 1;
 						}
-						elsif($opcode == TFTP_OPCODE_ERROR)
+						elsif ($opcode == TFTP_OPCODE_ERROR)
 						{
 							# message is ERR
 							$LASTERROR = sprintf "TFTP error message: %s", $datain;
-							return(undef);
+							return (undef);
 						}
 						else
 						{
 							# other messages...
 							$LASTERROR = sprintf "Opcode %d not supported as a reply to OACK\n", $opcode;
-							return(undef);
+							return (undef);
 						}
 					}
 					else
@@ -1055,17 +1211,17 @@ sub sendOACK
 					$retry++;
 				}
 			}
-			elsif($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_WRQ)
+			elsif ($self->{'_REQUEST_'}{'OPCODE'} eq TFTP_OPCODE_WRQ)
 			{
 				# opcode is WRQ
-				return(1);
+				return (1);
 			}
 		}
 		else
 		{
 			$! = $udpserver->sockopt(SO_ERROR);
 			$LASTERROR = sprintf "Socket SEND error: %s\n", $!;
-			return(undef);
+			return (undef);
 		}
 	}
 }
@@ -1077,26 +1233,34 @@ sub sendOACK
 sub sendERR
 {
 	my $self = shift;
-	my($errcode, $errmsg) = @_;
-	$errmsg or $errmsg = '';
+	my ($errcode, $errmsg) = @_;
+	# modified for supporting NETASCII transfers on 25/05/2009
+	#$errmsg or $errmsg = '';
+	$errmsg or $errmsg = $ERRORS{$errcode};
 
 	my $udpserver = $self->{'_UDPSERVER_'};
 
-	if($udpserver->send(pack("nnZ*", 5, $errcode, $errmsg)))
+	if ($udpserver->send(pack("nnZ*", 5, $errcode, $errmsg)))
 	{
-		return(1);
+		return (1);
 	}
 	else
 	{
 		$! = $udpserver->sockopt(SO_ERROR);
 		$LASTERROR = sprintf "Socket SEND error: %s\n", $!;
-		return(undef);
+		return (undef);
 	}
 }
 
+sub server
+{
+	my $self = shift;
+	return $self->{'_UDPSERVER_'};
+}
+
 sub error
 {
-	return($LASTERROR);
+	return ($LASTERROR);
 }
 
 # Preloaded methods go here.
@@ -1124,6 +1288,8 @@ Net::TFTPd - Perl extension for Trivial File Transfer Protocol Server
   $tftpRQ->processRQ()
     or die "Error processing TFTP request: %s", Net::TFTPd->error;
 
+  printf "%u bytes has been transferred", $tftpRQ->getTotalBytes() || 0;
+
 =head1 DESCRIPTION
 
 C<Net::TFTPd> is a class implementing a simple I<Trivial File Transfer Protocol> server in Perl as described in RFC1350.
@@ -1184,6 +1350,14 @@ Valid options are:
   BlkSize    Minimum blocksize to negotiate for transfers           512
   CallBack   Reference to code executed for each transferred block    -
   Debug      Activates debug mode (verbose)                           0
+  Family     Address family IPv4/IPv6                              IPv4
+               Valid values for IPv4:
+                 4, v4, ip4, ipv4, AF_INET (constant)
+               Valid values for IPv6:
+                 6, v6, ip6, ipv6, AF_INET6 (constant)
+
+B<NOTE>:  IPv6 requires B<IO::Socket::IP>.  Failback is B<IO::Socket::INET> 
+and only IPv4 support.
 
 =head2 CallBack
 
@@ -1221,6 +1395,53 @@ When the method returns, the program should fork() and process the request invok
 
 Processes a request and returns 1 if success, undef if error.
 
+=head2 getFileName()
+
+  $ret = $request->getFileName();
+
+Returns the requested file name.
+
+=head2 getMode()
+
+  $ret = $request->getMode();
+
+Returns the transfer mode for the request.
+
+=head2 getBlkSize()
+
+  $ret = $request->getBlkSize();
+
+Returns the block size used for the transfer.
+
+=head2 server()
+
+  $ret = $request->server();
+
+Return B<IO::Socket::*> object for the created server.  
+All B<IO::Socket::*> accessors can then be called.
+
+=head2 getPeerAddr()
+
+  $ret = $request->getPeerAddr();
+
+Returns the address of the requesting client.
+
+=head2 getPeerPort()
+
+  $ret = $request->getPeerMode();
+
+Returns the port of the requesting client.
+
+=head2 getTotalBytes()
+
+  $ret = $request->getTotalBytes();
+
+Returns the number of bytes transferred for the request.
+
+=head1 CREDITS
+
+Thanks to Michael Vincent (E<lt>VINSWORLDE<gt>) for the NETASCII support, transferred bytes and IPv6 patches.
+
 =head1 AUTHOR
 
 Luigino Masarati, E<lt>lmasarati@hotmail.comE<gt>
@@ -1230,3 +1451,4 @@ Luigino Masarati, E<lt>lmasarati@hotmail.comE<gt>
 L<Net::TFTP>.
 
 =cut
+
